#define  PIN_RESET   PIN_C0
#define  PIN_CD      PIN_C4
#define  PIN_RD      PIN_C2
#define  PIN_WR      PIN_C1
#define  PIN_CE      PIN_C3

#define  PIN_SHOOT   PIN_B5
#define  PIN_LEFT    PIN_B6
#define  PIN_RIGHT   PIN_B7

#define  RDWR_DELAY  20

#define  READY_FOR_DATA    0x03
#define  READY_FOR_COMMAND 0x03

#define  X_LIMIT     239
#define  Y_LIMIT     127

#define  TEXT_ADDRESS      0x1400
#define  GRAPH_ADDRESS     0x0000
#define  CHARS_PER_LINE    30
#define  LINES             16

#define  SHOTS_POSITION    ( TEXT_ADDRESS + 10 )
#define  INVADERS_POSITION ( TEXT_ADDRESS + 25 )

#define  OR_MODE           0x00
#define  EXOR_MODE         0x01
#define  AND_MODE          0x02

#define  GRAPH             0x08
#define  TEXT              0x04
#define  CURSOR            0x02
#define  BLINK             0x01

#define  X_PLUS            0x01
#define  Y_PLUS            0x10

#define  X_ERROR           0x01
#define  Y_ERROR           0x02

#define  CANNON_POSITION    ( CHARS_PER_LINE * 116 )

#define  CANNON_HEIGHT     10
#define  MISSILE_HEIGHT    8
#define  UFO_HEIGHT        8
#define  UFO_WIDTH         24
#define  MISSILE_WIDTH     8



#include "C:\Program Files\PICC\Projects\PIC16_EXP\PIC16_EXP.h"
#include "STDLIB.H"

const int8 cannon[ CANNON_HEIGHT ][ 3 ] = {  0x00,0x3C,0x00,
                           0x00,0x24,0x00,
                           0x3E,0x24,0x7C,
                           0xE2,0xC3,0x47,
                           0x02,0x81,0x40,
                           0xFE,0x81,0x7F,
                           0x22,0x42,0x44,
                           0x22,0x24,0x44,
                           0x3C,0x18,0x3C,
                           0xC0,0xFF,0x03 };
                        
const int8 missile[ MISSILE_HEIGHT ] = { 0x10,0x38,0x28,0x6C,0xAA,0xBA,0xC6,0x00 };

const int8 ufo[ UFO_HEIGHT ][ 3 ] = {  0x00,0x7E,0x00,
                           0xF8,0xA5,0x1F,
                           0x14,0x5A,0x28,
                           0xE2,0x5A,0x47,
                           0x14,0x5A,0x28,
                           0xF8,0xA5,0x1F,
                           0x00,0x7E,0x00,
                           0x00,0x00,0x00 };
                           
const char _invaders[] = { "Killed -" };
const char _shots[]    = { "Shots -" };

                          

int8 indx_y, count, coord_x_fine, coord_x_approx;
char shift_buffer[ 4 ];

unsigned int8 cannon_coord;

int8 mis_coord_x , mis_coord_y, missile_exists;

int8 ufo_coord_x, ufo_coord_y, ufo_direction, ufo_count;

int16 address ;

int8  crash;

char shots[ 2 ];
char invaders[ 2 ];

 void init_ports ( void )
//   
{
   set_tris_d ( 0x00 );
   set_tris_c ( 0x00 );
   set_tris_b ( 0xF0 );
     
   output_high ( PIN_WR );
   output_high ( PIN_RD );
   output_high ( PIN_RESET );
   output_high ( PIN_CD );
   output_high ( PIN_CE );
  
}



void reset_lcd ( void )
//    
{
   //     
   #ifndef PROTEUS 
   output_low( PIN_RESET );
   delay_ms( 10 );
   output_high( PIN_RESET );
   delay_ms( 10 );
   #endif
   
   #ifdef PROTEUS
   output_low( PIN_RESET );
   delay_cycles ( 10 );
   output_high( PIN_RESET );
   delay_cycles ( 10 );
   #endif
}

  void write_data ( char ch )
{
 //    
   output_d ( ch );
   output_low ( PIN_CD );
   output_low ( PIN_CE );
   output_low ( PIN_WR );
   delay_cycles ( RDWR_DELAY );
   output_high ( PIN_WR );
   output_high ( PIN_CE );
}

  void write_command ( char ch )
{
 //    
   output_d ( ch );
   output_high ( PIN_CD );
   output_low ( PIN_CE );
   output_low ( PIN_WR );
   delay_cycles ( RDWR_DELAY );
   output_high ( PIN_WR );
   output_high ( PIN_CE );
   output_low   ( PIN_CD );
}

  char read_data ( void )
{
 //    
   char ch;
   
   set_tris_d ( 0xFF );
   output_low ( PIN_CD );
   output_low ( PIN_CE );
   output_low ( PIN_RD );
   delay_cycles ( RDWR_DELAY );
   ch = input_b ();
   output_high ( PIN_RD );
   output_high ( PIN_CE );
   set_tris_d  ( 0x00 );
   return ( ch );
}
  char get_status ( void )
{
 //   
   char ch;
   
   set_tris_d ( 0xFF );
   output_high( PIN_CD );
   output_low ( PIN_CE );
   output_low ( PIN_RD );
   delay_cycles ( RDWR_DELAY );
   ch = input_d ();
   output_high ( PIN_RD );
   output_high ( PIN_CE );
   output_low  ( PIN_CD );
   set_tris_d  ( 0x00 );
   return ( ch );
}


 void set_data ( char ch )
{
 //     
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   while ( ( get_status() & READY_FOR_COMMAND ) != 0x03 )
   { ;;; }
   write_command ( 0xC4 );
}

 void set_data_incr ( char ch )
{
 //      c   
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   while ( ( get_status() & READY_FOR_COMMAND ) != 0x03 )
   { ;;; }
   write_command ( 0xC0 );
}

 void set_data_decr ( char ch )
{
 //      c   
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   while ( ( get_status() & READY_FOR_COMMAND ) != 0x03 )
   { ;;; }
   write_command ( 0xC2 );
}



void set_graph_text_on_off ( char ch )
{
 // ???       
   while ( ( get_status() & READY_FOR_COMMAND ) != 0x03 )
   { ;;; }
   write_command ( 0x90 | ( ch & 0x0F ));
}

void set_graph_text_mode ( char ch )
{
 // ???      
   while ( ( get_status() & READY_FOR_COMMAND ) != 0x03 )
   { ;;; }
   write_command ( 0x80 | ( ch & 0x0F ));
}

void set_graph_address ( int16 addr )
{
 // ???     
   char ch;
   
   ch = addr & 0x00FF;
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   
   ch = addr >> 8;
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   
   while ( ( get_status() & READY_FOR_COMMAND ) != 0x03 )
   { ;;; }
   write_command ( 0x42 );   
}

void set_text_address ( int16 addr )
{
 // ???    
   char ch;
   
   ch = addr & 0x00FF;
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   
   ch = addr >> 8;
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   
   while ( ( get_status() & READY_FOR_COMMAND ) != 0x03 )
   { ;;; }
   write_command ( 0x40 );   
}

void set_graph_area ( int16 addr )
{
 // ???   ,     
   char ch;
   
   ch = addr & 0x00FF;
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   
   ch = addr >> 8;
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   
   while ( ( get_status() & READY_FOR_COMMAND ) != 0x03 )
   { ;;; }
   write_command ( 0x43 );   
}

void set_text_area ( int16 addr )
{
 // ???   ,       
   char ch;
   
   ch = addr & 0x00FF;
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   
   ch = addr >> 8;
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   
   while ( ( get_status() & READY_FOR_COMMAND ) != 0x03 )
   { ;;; }
   write_command ( 0x41 );   
}

void set_address_pointer ( int16 addr )
{
 // ???        
   char ch;
   
   ch = addr & 0x00FF;
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   
   ch = addr >> 8;
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_data ( ch );
   
   while ( ( get_status() & READY_FOR_COMMAND ) != 0x03 )
   { ;;; }
   write_command ( 0x24 );   
}

void set_auto_write_mode ( void  )
{
 // ???     
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_command ( 0xB0 );
}

void set_auto_read_mode ( void  )
{
 // ???     
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_command ( 0xB1 );
}

void reset_auto_mode ( void  )
{
 // ???   
   while ( ( get_status() & READY_FOR_DATA ) != 0x03 )
   { ;;; }
   write_command (  0xB2 );
}

void draw_cannon ( int8 cannon_coord )
{
   output_high( PIN_B4 );
   
   for ( indx_y = 0; indx_y < CANNON_HEIGHT; indx_y++ )
   {
      shift_buffer[ 0 ] = 0x00;
      shift_buffer[ 1 ] = cannon[ indx_y ][ 0 ];
      shift_buffer[ 2 ] = cannon[ indx_y ][ 1 ];
      shift_buffer[ 3 ] = cannon[ indx_y ][ 2 ] ;
         
      // ,    
      coord_x_fine = cannon_coord & 0x07;
      coord_x_approx = cannon_coord >> 3;
      
      if ( coord_x_fine != 0 )
      {
         for ( count = 0 ; count < coord_x_fine ; count++ )
               rotate_right( shift_buffer , 4 );
      }            
   
      address = indx_y;
      address = address * CHARS_PER_LINE;
      address = address + coord_x_approx + CANNON_POSITION;
   
      set_address_pointer( address );
   
      for ( count = 3 ; count != 0xFF ; count-- )
      {
         set_data_incr( shift_buffer[ count ] );
      }
   }
   output_low( PIN_B4 );
}

void draw_missile ( int8 missile_coord_x, int8 missile_coord_y )
{
   output_high( PIN_B3 );
   
   for ( indx_y = 0; indx_y < MISSILE_HEIGHT; indx_y++ )
   {
      shift_buffer[ 0 ] = 0x00;
      shift_buffer[ 1 ] = missile[ indx_y ];

         
      // ,    
      coord_x_fine = missile_coord_x & 0x07;
      coord_x_approx = missile_coord_x >> 3;
      
      if ( coord_x_fine != 0 )
      {
         for ( count = 0 ; count < coord_x_fine ; count++ )
               rotate_right( shift_buffer , 2 );
      }            
   
      address = indx_y + missile_coord_y;
      address = address * CHARS_PER_LINE;
      address = address + coord_x_approx ;
   
      set_address_pointer( address );
   
      for ( count = 1 ; count != 0xFF ; count-- )
      {
         set_data_incr( shift_buffer[ count ] );
      }
   }
   output_low( PIN_B3 );
}

 void destroy_missile( int8 missile_coord_x, int8 missile_coord_y )
 {
         //       
         missile_exists = 0x00;
         
         for ( indx_y = 0; indx_y < MISSILE_HEIGHT; indx_y++ )
         {
            coord_x_approx = missile_coord_x >> 3;
            
            address = indx_y + missile_coord_y ;
            address = address * CHARS_PER_LINE;
            address = address + coord_x_approx ;
   
             set_address_pointer( address );
   
            for ( count = 1 ; count != 0xFF ; count-- )
            {
               set_data_incr( 0x00 );
            }
         }      
 }


void draw_ufo ( int8 UFObj_coord_x, int8 UFObj_coord_y )
{
   //output_high( PIN_B2 );
   
   for ( indx_y = 0; indx_y < UFO_HEIGHT; indx_y++ )
   {
      shift_buffer[ 0 ] = 0x00;
      shift_buffer[ 1 ] = ufo[ indx_y ][ 0 ];
      shift_buffer[ 2 ] = ufo[ indx_y ][ 1 ]; 
      shift_buffer[ 3 ] = ufo[ indx_y ][ 2 ];
         
      // ,    
      coord_x_fine = UFObj_coord_x & 0x07;
      coord_x_approx = UFObj_coord_x >> 3;
      
      if ( coord_x_fine != 0 )
      {
         for ( count = 0 ; count < coord_x_fine ; count++ )
               rotate_right( shift_buffer , 4 );
      }            
   
      address = indx_y + UFObj_coord_y;
      address = address * CHARS_PER_LINE;
      address = address + coord_x_approx ;
      
   
      set_address_pointer( address );
   
      for ( count = 3 ; count != 0xFF ; count-- )
      {
         set_data_incr( shift_buffer[ count ] );
      }
   }
  // output_low( PIN_B2 );
}

void destroy_ufo ( int8 UFObj_coord_x, int8 UFObj_coord_y )
{
   //output_high( PIN_B2 );
   
   for ( indx_y = 0; indx_y < UFO_HEIGHT; indx_y++ )
   {

      coord_x_approx = UFObj_coord_x >> 3;      
   
      address = indx_y + UFObj_coord_y;
      address = address * CHARS_PER_LINE;
      address = address + coord_x_approx ;
      
   
      set_address_pointer( address );
   
      for ( count = 3 ; count != 0xFF ; count-- )
      {
         set_data_incr( 0x00 );
      }
   }
  // output_low( PIN_B2 );
}


/*
void text_put_ROM_string ( const char ing )
{
   while ( *string != 0x00 )
   {
      set_data_incr ( *string - 0x20 );
      string++;   
   }
}*/

void init_data ( void )
//   
{
   cannon_coord = 0;
   missile_exists = 0;
   draw_cannon ( cannon_coord );
   
   
   ufo_count = rand();
   ufo_count = ( ufo_count >> 4 ) + 0x08;
   
   ufo_coord_y = 2 + ( 0x07 & rand ( ));
   ufo_coord_x = 50 + ( 0x7F & rand());
   
   ufo_direction = ( rand() & 0x01 );
   crash = 0x00;
   
   shots[ 0 ] = 0x10;
   shots[ 1 ] = 0x10;
   
   invaders[ 0 ] = 0x10;
   invaders[ 1 ] = 0x10;
   
    
}

void incr_invaders ( void )
{
   invaders [ 0 ] = invaders[ 0 ] + 1;
   if ( invaders[ 0 ] >= 0x1A )
   {
      invaders[ 0 ] = 0x10;
      invaders[ 1 ] = invaders[ 1 ] + 1;
         if( invaders [ 1 ] >= 0x1A ) 
            invaders[ 1 ] = 0x10;
              
   }
   set_address_pointer ( INVADERS_POSITION );
   set_data_incr ( invaders [ 1 ]);
   set_data_incr ( invaders [ 0 ]);


}

void incr_shots ( void )
{
   shots [ 0 ] = shots [ 0 ] + 1;
   if ( shots[ 0 ] >= 0x1A )
   {
      shots[ 0 ] = 0x10;
      shots[ 1 ] = shots[ 1 ] + 1;
         if( shots [ 1 ] >= 0x1A ) 
            shots[ 1 ] = 0x10;
              
   }
   set_address_pointer ( SHOTS_POSITION );
   set_data_incr ( shots [ 1 ]);
   set_data_incr ( shots [ 0 ]);


}


#int_RTCC
RTCC_isr() 
{
  
   
   //    ""
   if ( !( input( PIN_LEFT )) & input ( PIN_RIGHT )) 
   {  
      if ( cannon_coord > 0 )
      {
         cannon_coord--;
      }   
   }
   
   //    ""
   if ( !( input( PIN_RIGHT )) & input ( PIN_LEFT )) 
   {  
      if ( cannon_coord < 212  )
      {
         cannon_coord++;
      } 
   }
   
   //   
   draw_cannon( cannon_coord );
   
   
   //     
   if ( !missile_exists )
   {
   // ,    ""
   if  ( !input ( PIN_SHOOT ) & ( crash == 0x00 )  )
      {
         missile_exists = 0xFF;
         mis_coord_x = cannon_coord + 9;
         mis_coord_y = 116 - 12 ;
         incr_shots();
      }
   }
   //    
   else
   {
      //     
      mis_coord_y--;
      if ( mis_coord_y == 0x00 )
      {
         destroy_missile( mis_coord_x, mis_coord_y );
         
      }
      else
      {
          // ,      
         if (( mis_coord_y > ufo_coord_y ) & ( mis_coord_y < ( ufo_coord_y + UFO_HEIGHT )) &
         ( mis_coord_x > ufo_coord_x ) & ( mis_coord_x < ( ufo_coord_x + UFO_WIDTH ) ))
         {
            crash = 0xFF;
            ufo_count = 50;
         }
         if ( crash != 0xFF )
         {
            if (( ( mis_coord_y + MISSILE_HEIGHT ) > ufo_coord_y ) & (( mis_coord_y + MISSILE_HEIGHT ) < ( ufo_coord_y + UFO_HEIGHT )) &
               ( mis_coord_x > ufo_coord_x ) & ( mis_coord_x < ( ufo_coord_x + UFO_WIDTH ) ))
            {
               crash = 0xFF;
               ufo_count = 50;
            }
         }
         
         
         if ( crash != 0xFF )
         {
            if ((  mis_coord_y  > ufo_coord_y ) & ( mis_coord_y   < ( ufo_coord_y + UFO_HEIGHT )) &
               ( ( mis_coord_x + MISSILE_WIDTH ) > ufo_coord_x ) & ( ( mis_coord_x+ MISSILE_WIDTH ) < ( ufo_coord_x + UFO_WIDTH ) ))
            {
                  crash = 0xFF;
                  ufo_count = 50;
            }
         }
         
         if ( crash != 0xFF )
         {
            if ((  ( mis_coord_y + MISSILE_HEIGHT ) > ufo_coord_y ) & ( ( mis_coord_y +MISSILE_HEIGHT )  < ( ufo_coord_y + UFO_HEIGHT )) &
               ( ( mis_coord_x + MISSILE_WIDTH ) > ufo_coord_x ) & ( ( mis_coord_x+ MISSILE_WIDTH ) < ( ufo_coord_x + UFO_WIDTH ) ))
            {
               crash = 0xFF;
               ufo_count = 50;
            }
         }
         
      }
   }

   if ( crash == 0x00 )
   {   //       
   ufo_count--;
      if ( ufo_count == 0x00 )
      {
        ufo_count = rand();
        ufo_count = ( ufo_count >> 4 ) + 0x20;
        
        ufo_direction = ( 0x01 & rand() );
      }     
      if ( ufo_direction == 0x00 )
      {
            ufo_coord_x++;
            if ( ufo_coord_x > 210 )
            {
               ufo_direction = 0x01;
               ufo_count += 0x06;
               ufo_coord_x--;
            }
      }
      else 
      {
         ufo_coord_x--;
            if ( ufo_coord_x == 0xFF )
            {
               ufo_direction = 0x00;
               ufo_count += 0x06;
               ufo_coord_x++; 
            }
      }
      draw_ufo( ufo_coord_x, ufo_coord_y );
      if ( missile_exists )
          draw_missile ( mis_coord_x, mis_coord_y );
   }
   else
   {
      destroy_ufo ( ufo_coord_x, ufo_coord_y );
      destroy_missile( mis_coord_x, mis_coord_y );
      ufo_count--;
      if ( ufo_count == 0x00 )
      {
         crash = 0x00; 
         
         ufo_count = rand();
         ufo_count = ( ufo_count >> 4 ) + 0x20;
         ufo_direction = ( 0x01 & rand() );
         
         ufo_coord_y = 2 + ( 0x1F & rand());
         ufo_coord_x = 50 + ( 0x7F & rand());
         
         incr_invaders();
         
         
         
     }
   }
   
  }




void main()
{

   setup_adc_ports(NO_ANALOGS);
   setup_adc(ADC_OFF);
   setup_psp(PSP_DISABLED);
   setup_spi(FALSE);
   setup_timer_0(RTCC_INTERNAL|RTCC_DIV_128);
   setup_timer_1(T1_DISABLED);
   setup_timer_2(T2_DISABLED,0,1);
   enable_interrupts(INT_RTCC);
   

   // TODO: USER CODE!!
   
  init_ports();
  reset_lcd();
  reset_auto_mode();
  
  set_text_address ( TEXT_ADDRESS );
  set_graph_address ( GRAPH_ADDRESS );
  set_graph_area( CHARS_PER_LINE );
  set_text_area( CHARS_PER_LINE );
  set_graph_text_on_off( GRAPH | TEXT );
  set_graph_text_mode( OR_MODE );
  
  init_data();
  
  enable_interrupts(GLOBAL);
   
  
 while (true )
 {
   ;;
 }

}
